<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 1/25/16
 * Time: 10:00 PM
 */

namespace IMATHUZH\Qfq\Core;

use IMATHUZH\Qfq\Core\Database\Database;
use IMATHUZH\Qfq\Core\Form\Form;
use IMATHUZH\Qfq\Core\Form\FormAsFile;
use IMATHUZH\Qfq\Core\Helper\OnArray;
use IMATHUZH\Qfq\Core\Renderer\BaseRenderer;
use IMATHUZH\Qfq\Core\Report\Link;
use IMATHUZH\Qfq\Core\Store\Store;

/**
 * Class BuildFormBootstrap
 * @package qfq
 */
class FormBuilder {

    /**
     * @var array
     */
    protected $formSpec = array();  // copy of the loaded form
    /**
     * @var array
     */
    protected $feSpecAction = array(); // copy of all formElement.class='action' of the loaded form
    /**
     * @var array
     */
    protected $feSpecNative = array(); // copy of all formElement.class='native' of the loaded form
    /**
     * @var Store
     */
    protected $store = null;
    /**
     * @var Link
     */
    protected $link = null;
    /**
     * @var Database[] - Array of Database instantiated class
     */
    protected $dbArray = array();
    /**
     * @var bool|mixed
     */
    protected $dbIndexData = false;
    /**
     * @var bool|string
     */
    protected $dbIndexQfq = false;


    /**
     * @param array $formSpec
     * @param array $feSpecAction
     * @param array $feSpecNative
     * @param array $db Array of 'Database' instances
     * @throws \CodeException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public function __construct(array $formSpec, array $feSpecAction, array $feSpecNative, array $db) {
        $this->formSpec = $formSpec;
        $this->feSpecAction = $feSpecAction;
        $this->feSpecNative = $feSpecNative;
        $this->store = Store::getInstance();
//        $this->dbIndexData = $this->store->getVar(SYSTEM_DB_INDEX_DATA, STORE_SYSTEM);
        $this->dbIndexData = $formSpec[F_DB_INDEX];
        $this->dbIndexQfq = $this->store->getVar(SYSTEM_DB_INDEX_QFQ, STORE_SYSTEM);

        $this->dbArray = $db;
        //$this->evaluate = new Evaluate($this->store, $this->dbArray[$this->dbIndexData]);
        // $this->showDebugInfoFlag = Support::findInSet(SYSTEM_SHOW_DEBUG_INFO_YES, $this->store->getVar(SYSTEM_SHOW_DEBUG_INFO, STORE_SYSTEM));

        $this->sip = $this->store->getSipInstance();

        $this->link = new Link($this->sip, $this->dbIndexData,'', $formSpec);


        $this->buildRowName = [
            FE_TYPE_CHECKBOX => 'Native',
            FE_TYPE_DATE => 'Native',
            FE_TYPE_DATETIME => 'Native',
            'dateJQW' => 'Native',
            'datetimeJQW' => 'Native',
            'email' => 'Native',
            'gridJQW' => 'Native',
            FE_TYPE_EXTRA => 'Native',
            FE_TYPE_TEXT => 'Native',
            FE_TYPE_EDITOR => 'Native',
            FE_TYPE_TIME => 'Native',
            FE_TYPE_NOTE => 'Native',
            FE_TYPE_PASSWORD => 'Native',
            FE_TYPE_RADIO => 'Native',
            FE_TYPE_SELECT => 'Native',
            FE_TYPE_SUBRECORD => 'Subrecord',
            FE_TYPE_UPLOAD => 'Native',
            FE_TYPE_ANNOTATE => 'Native',
            FE_TYPE_IMAGE_CUT => 'Native',
            FE_TYPE_CHAT => 'Native',
            'fieldset' => 'Fieldset',
            'pill' => 'Pill',
            'templateGroup' => 'TemplateGroup',
        ];

        //$this->symbol[SYMBOL_DELETE] = "<span class='glyphicon " . GLYPH_ICON_DELETE . "'></span>";

        $this->ttContentUid = $this->store->getVar(TYPO3_TT_CONTENT_UID, STORE_TYPO3);
        $this->formId = $this->formSpec[F_ID] ?? 0;
        $this->recordId = $this->store->getVar(CLIENT_RECORD_ID, STORE_SIP . STORE_RECORD . STORE_TYPO3 . STORE_ZERO);
    }


    /**
     * Builds complete 'form'.
     *
     * @param string $mode FORM_LOAD | FORM_UPDATE | FORM_SAVE
     *
     * @param bool $htmlElementNameIdZero
     * @param array $latestFeSpecNative
     * @return string|array   $mode=LOAD_FORM: The whole form as HTML, $mode=FORM_UPDATE /FORM_SAVE: array of all
     *                        formElement.dynamicUpdate-yes  values/states
     * @throws \CodeException
     * @throws \DbException
     * @throws \DownloadException
     * @throws \PhpOffice\PhpSpreadsheet\Exception
     * @throws \PhpOffice\PhpSpreadsheet\Reader\Exception
     * @throws \PhpOffice\PhpSpreadsheet\Writer\Exception
     * @throws \Twig\Error\LoaderError
     * @throws \Twig\Error\RuntimeError
     * @throws \Twig\Error\SyntaxError
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public function process($mode, $htmlElementNameIdZero = false, $latestFeSpecNative = array()) {

        $this->formSpec[FORM_ELEMENTS_NATIVE] = $this->feSpecNative; // Container Elements are included here
        $this->formSpec[FORM_ELEMENTS_ACTION] = $this->feSpecAction;

        // FE Types that must not be processed
        $skipFeTypes = [];
        if($mode == FORM_UPDATE){ $skipFeTypes = [FE_TYPE_ANNOTATE];}
        else if($mode == FORM_SAVE){ $skipFeTypes = [FE_TYPE_ANNOTATE];}

        // Remove elements that should be skipped
        $formSpec = $this->formSpec;
        for($i = count($formSpec[FORM_ELEMENTS_NATIVE]) - 1; $i >=0 ; $i--){
            if(in_array($formSpec[FORM_ELEMENTS_NATIVE][$i][FE_TYPE], $skipFeTypes)){
                unset($formSpec[FORM_ELEMENTS_NATIVE][$i]);
            } else {
                switch ($mode) {
                    case FORM_UPDATE:
                        // Render only Pills or FEs that are set to 'dynamic update'
                        if((!isset($formSpec[FORM_ELEMENTS_NATIVE][$i][FE_DYNAMIC_UPDATE]) || $formSpec[FORM_ELEMENTS_NATIVE][$i][FE_DYNAMIC_UPDATE] != 'yes') && $formSpec[FORM_ELEMENTS_NATIVE][$i][FE_TYPE] != FE_TYPE_PILL){
                            unset($formSpec[FORM_ELEMENTS_NATIVE][$i]);
                        }
                        break;
                    case FORM_SAVE:
                        // Remove FEs that are part of a template group
                        // TODO: Find better solution for str_contains
                        if($formSpec[FORM_ELEMENTS_NATIVE][$i][FE_TG_INDEX] != 0 || str_contains($formSpec[FORM_ELEMENTS_NATIVE][$i][FE_NAME], '%')) {
                            unset($formSpec[FORM_ELEMENTS_NATIVE][$i]);
                        }
                        break;
                    default:
                        break;
                }
            }
        }
        $formSpec[FORM_ELEMENTS_NATIVE] = array_values($formSpec[FORM_ELEMENTS_NATIVE]); // re-index the array

        // Process depending on Mode
        switch ($mode){
            case FORM_LOAD:
                $form = new Form($this->formSpec);
                $renderer = BaseRenderer::getInstance();
                return $renderer->renderFormHtml($form, RENDER_MODE_HTML);

            case FORM_UPDATE:
                $formSpec[FORM_ELEMENTS_NATIVE] = array_values($formSpec[FORM_ELEMENTS_NATIVE]); // re-index the array

                $form = new Form($formSpec);
                $renderer = BaseRenderer::getInstance();
                return $renderer->renderFormJson($form);

            case FORM_SAVE:
                // The form has already been saved at this point by QuickFormQuery->saveForm!
                // If the record was updated (r!=0), then the form is updated via json - like in FORM_UPDATE!
                // If the record was inserted (r=0), then a FORM_LOAD is performed
                $recordId = $this->store->getVar(SIP_RECORD_ID, STORE_SIP . STORE_ZERO);
                if ($mode === FORM_SAVE && $recordId != 0) {
                    $form = new Form($formSpec);
                    $renderer = BaseRenderer::getInstance();
                    $arrayForJson = $renderer->renderFormJson($form);
                    // element-update: with 'value'
                    $md5 = $this->buildRecordHashMd5($this->formSpec[F_TABLE_NAME], $recordId, $formSpec[F_PRIMARY_KEY]);
                    $arrayForJson[][API_ELEMENT_UPDATE][DIRTY_RECORD_HASH_MD5][API_ELEMENT_ATTRIBUTE]['value'] = $md5;
                    return $arrayForJson;
                }
                return array(); // Save with id=0 results in FORM_LOAD - no output needed here
            default:
                break;
        }
    }

    /**
     * @param $tableName
     * @param $recordId
     * @param string $primaryKey
     *
     * @return string
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public function buildRecordHashMd5($tableName, $recordId, $primaryKey = F_PRIMARY_KEY_DEFAULT) {
        $record = array();

        if ($recordId != 0) {
            $record = $this->dbArray[$this->dbIndexData]->sql("SELECT * FROM `$tableName` WHERE `$primaryKey`=?", ROW_EXPECT_1, [$recordId], "Record to load not found. "
                . (FEATURE_FORM_FILE_SYNC ? FormAsFile::errorHintFormImport($tableName) : ''));
        }

        if (isset($record[F_FILE_STATS])) {
            // why: The column "fileStats" in the Form table is modified when a form is exported to a file but nothing else changes.
            unset($record[F_FILE_STATS]);
        }

        return OnArray::getMd5($record);
    }

    // TODO: Check if this is rly needed
    /**
     * @return string
     */
    private function getTabId() {
        return 'qfqTabs-' . $this->store->getVar(TYPO3_TT_CONTENT_UID, STORE_TYPO3) . '-'
            . $this->formSpec[F_ID] . '-' . $this->store->getVar(CLIENT_RECORD_ID, STORE_TYPO3 . STORE_SIP . STORE_RECORD . STORE_ZERO);
    }

    // TODO: Add Jay's IPA functionality to multiform
    /*protected function buildMultiForm($filter, $modeCollectFe, array &$rcJson) {

        $htmlElements = '';
        $rcJson = array();
        $flagTable = true;

        // Author: Zhoujie Li
        // if parameter multiFormAddRow === 1 then true else false
        $addRow =  isset($this->formSpec[F_MULTIFORM_ADD_ROW]) && $this->formSpec[F_MULTIFORM_ADD_ROW] == '1';
        // if parameter multiFormDeleteRow === 1 then true else false
        $deleteRow = isset($this->formSpec[F_MULTIFORM_DELETE_ROW]) && $this->formSpec[F_MULTIFORM_DELETE_ROW] == '1';
        $finalDeleteBtn = '';
        // End from author

        $counter = 0;

        if (!isset($this->formSpec[F_MULTI_WRAP])) {
            $this->formSpec[F_MULTI_WRAP] = '';
        }
        if (!isset($this->formSpec[F_MULTI_CELL_WRAP])) {
            $this->formSpec[F_MULTI_CELL_WRAP] = '';
        }
        if ($this->formSpec[F_MULTI_WRAP] == '') {
            $this->formSpec[F_MULTI_WRAP] = '<table class="table table-multi-form qfq-table-100">';
            $this->formSpec[F_MULTI_CELL_WRAP] = '<td>';
        } else {
            $flagTable = false;
            if ($this->formSpec[F_MULTI_CELL_WRAP] == '') {
                $this->formSpec[F_MULTI_CELL_WRAP] = '<div>';
            }
        }

        $parentRecords = $this->evaluate->parse($this->formSpec[F_MULTI_SQL], ROW_REGULAR);
        // Check for 'id' or '_id' as column name
        $idName = isset($parentRecords[0]['_' . F_MULTI_COL_ID]) ? '_' . F_MULTI_COL_ID : F_MULTI_COL_ID;

        // Author: Zhoujie Li
        // if addRow is true add 1 additional element to $parentRecord.
        if($addRow){
            HelperFormElement::addGenericElements($parentRecords, 1, $idName);
        }
        // Get get the Total records need to checkt the last element with id 0
        $totalRecords = count($parentRecords);
        // count the current index
        $currentRecordIndex = 0;


        // If form.parameter.processRow is set, checkboxes are added to parentRecords
        if (isset($this->formSpec[F_PROCESS_ROW])) {
            // check or uncheck processRow
            $this->processRow($parentRecords, $idName, $counter);
        }
        // End from author
//        // Important to destroy the reference (further usage)
//        unset ($array);


        // No rows: nothing to do.
        if (empty($parentRecords)) {
            return $this->formSpec[F_MULTI_MSG_NO_RECORD];
        }

        // Check that a column 'id' is given
        if (!isset($parentRecords[0][$idName])) {
            throw new \UserFormException(
                json_encode([ERROR_MESSAGE_TO_USER => 'Missing column "_' . F_MULTI_COL_ID . '"', ERROR_MESSAGE_TO_DEVELOPER => $this->formSpec[F_MULTI_SQL]]),
                ERROR_INVALID_OR_MISSING_PARAMETER);
        }

        // This is a dirty workaround for formSave: clear FORM STORE (already outdated values).
        // Otherwise those outdated values will be taken to fill non primary FE in multiForm (which are garbage).
        // Better solution would be to have FORM_STORE in sync.
        $this->store::unsetStore(STORE_FORM);

        $storeVarBase = $this->evaluate->parse($this->formSpec[FE_FILL_STORE_VAR]);
        if (isset($storeVarBase[0])) {
            // FE_FILL_STORE_VAR might return multiple rows (not useful), nevertheless the first row should become the only one.
            $storeVarBase = $storeVarBase[0];
        }
        if (!is_array($storeVarBase)) {
            $storeVarBase = array();
        }

        // Per row, iterate over all form elements
        foreach ($parentRecords as $row) {
            // increment currentRecordIndexs
            $currentRecordIndex++;
            // Always start with a clean STORE_VAR
            $this->store->setStore($storeVarBase, STORE_VAR, true);

            $this->store->setStore(OnArray::keyNameRemoveLeadingUnderscore($row), STORE_PARENT_RECORD, true);
            $this->store->setVar(F_MULTI_COL_ID, $row[$idName], STORE_PARENT_RECORD); // In case '_id' is used, both '_id' and 'id' should be accessible.

            // Main record to load in STORE_RECORD
            $record = $this->dbArray[$this->dbIndexData]->sql('SELECT * FROM `' . $this->formSpec[F_TABLE_NAME] . '` WHERE `id`= ?',
                ROW_EXPECT_0_1, [$row[$idName]]);
            $this->store->setStore($record, STORE_RECORD, true);

            $jsonTmp = array();
            $feTmp = $this->feSpecNative;

            $leftColumns = $this->buildMultiFormLeftColumns($row, $this->formSpec[F_MULTI_CELL_WRAP]);
            $rightInputs = $this->elements($row[$idName], $filter, 0, $jsonTmp, $modeCollectFe,
                false, STORE_USE_DEFAULT, FORM_LOAD, true);


            // Author: Zhoujie Li
            if($deleteRow){
                $disable = $row[$idName] === 0 ? '' : 'disabled';
                $deleteRender = '';

                $strDefault = TOKEN_ACTION_DELETE . ':' . TOKEN_ACTION_DELETE_AJAX
                    . '|' . TOKEN_GLYPH . ':' . GLYPH_ICON_DELETE
                    . '|' . TOKEN_SIP . '|' . TOKEN_BOOTSTRAP_BUTTON . $deleteRender
                    . '|' . TOKEN_URL_PARAM . ':form=' . $this->formSpec[F_NAME] . '&r=' . $row[$idName] . '&'. FLAG_MULTIFORM .'=true'
                    . '|' . TOKEN_TOOL_TIP . ':Delete';

                $deleteBtn = $row[$idName] === 0 ? $this->buildButtonCode(MULTIFORM_DELETE_ROW_BTN,'','delete row',GLYPH_ICON_DELETE , '' , '', 'btn btn-default ' . MULTIFORM_DELETE_ROW_BTN)  : $this->link->renderLink($strDefault);

                $finalDeleteBtn = '<td style="text-align: center;">' . $deleteBtn . '</td>';
            }
            if ($flagTable) {
                $recordClass = 'record';
                $useRecordClass = ($row[$idName] != 0) ? Support::doAttribute(HTML_ATTR_CLASS, $recordClass, true) : '';
                $htmlId = Support::doAttribute(HTML_ATTR_ID, $this->getFormId($this->ttContentUid, $this->formSpec[F_ID], $this->recordId) . '-' . $row[$idName]);
                // If it's the last record and it is empty also addRow is true, add 'dummy-row' class; otherwise, use a regular 'tr' tag
                $tRowTag = ($totalRecords === $currentRecordIndex && empty($record) && $addRow) ? 'class="dummy-row"' : $useRecordClass . $htmlId;

                $htmlElements .= Support::wrapTag("<tr $tRowTag>", $leftColumns . $rightInputs . $finalDeleteBtn);
            } else {
                $htmlElements .= $rightInputs;
            }

            // Clean for the next round
            $this->feSpecNative = $feTmp;
            $this->store::unsetStore(STORE_RECORD);

            $rcJson = array_merge($rcJson, $jsonTmp);
        }

        // Author: Zhoujie Li
        if ($flagTable) {
            $html = Support::wrapTag('<thead><tr>', $this->buildMultiFormTableHead($parentRecords[0] , $deleteRow)) .
                '<tbody>' . $htmlElements . '</tbody>';
            // End from author
        } else {
            $html = $htmlElements;
        }


        $multiFormTable = Support::wrapTag($this->formSpec[F_MULTI_WRAP], $html);

        // Author: Zhoujie Li
        // render Button if $addRow is true
        if($addRow){
            // append button to html
            $multiFormTable .= $this->buildButtonCode(MULTIFORM_ADD_ROW_BTN,'','add new row', GLYPH_ICON_NEW,'','',$this->formSpec[F_SAVE_BUTTON_CLASS] . ' addRowButton');
        }
        // End from author
        return $multiFormTable;
    }*/

    // TODO: NoteForm Button
    /**
     * Build note button in form.
     *
     * @return string
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     * @throws \UserReportException
     *
     */
    /*
    private function buildNoteButton() {

        $recordId = $this->recordId;

        // NoteItem enabled?
        if (!Support::findInSet(FORM_BUTTON_NOTE, $this->formSpec[F_SHOW_BUTTON])) {
            return '';
        }

        $formId = $this->formSpec[NOTE_ITEM_COLUMN_ID];
        $formName = $this->store->getVar(FORM_NAME_FORM, STORE_SIP);

        $xId = $this->store->getVar(SIP_RECORD_ID, STORE_SIP);

        // Get pIdUser. 'User Switch' should be handled in STORE_SYSTEM > fillStoreSystemBySqlRowVariable
        $pIdUser = $this->store->getVar(SYSTEM_PID_USER, STORE_SYSTEM . STORE_ZERO);

        $sql = "SELECT COUNT(*) AS cnt FROM " . NOTE_ITEM_TABLE . " AS nf WHERE nf.formId =" . $formId . " AND nf.xId=" . $recordId
            . " AND (nf.pIdCreator = " . $pIdUser . " OR nf.access= 'all read' OR nf.access = 'all write')" . " GROUP BY nf.xId";
        // Count the number of notes related to the form and record, accessible by the user or publicly ('all')
        $sqlResult = $this->dbArray[$this->dbIndexData]->sql($sql, ROW_EXPECT_0_1);
        $noteTotal = $sqlResult['cnt'] ?? 0;
        $disabled = ($xId == '0' || $formName == NOTE_ITEM_NAME) ? HTML_ATTR_HIDDEN : '';

        // Query to find notes that are not marked as done
        $sqlResult = $this->dbArray[$this->dbIndexData]->sql("SELECT COUNT(*) AS cnt FROM " . NOTE_ITEM_TABLE .
            " AS nf WHERE nf.formId = " . $formId . " AND nf.xId = " . $recordId . " AND nf.isDone = 'no' AND nf.hasToolbar = 'yes' GROUP BY formId",
            ROW_EXPECT_0_1);
        $noteTotalOpen = $sqlResult['cnt'] ?? 0;

        // Tooltip note button
        $toolTip = 'Notes';
        // Set the badge class based on whether there are notes not done.
        $badgeClass = 'badge';
        $style = '';
        if ($noteTotalOpen > 0 && $noteTotal != 0) {
            $badgeClass = 'qfq-badge qfq-badge-warning';
            $toolTip .= " (open: $noteTotalOpen, total: $noteTotal)";
        }

        $msg = ($noteTotalOpen > 0) ? $noteTotalOpen . ' / ' . $noteTotal : $noteTotal;
        $badge = "<span class='$badgeClass' $style>$msg</span>";

        // Debug Info
        if (Support::findInSet(SYSTEM_SHOW_DEBUG_INFO_YES, $this->store->getVar(SYSTEM_SHOW_DEBUG_INFO, STORE_SYSTEM))) {
            $toolTip .= PHP_EOL . '--' . PHP_EOL . 'xId: ' . $recordId . PHP_EOL . 'formName:' . $formName;
        }

        $pageSlug = $this->store->getVar(TYPO3_PAGE_SLUG, STORE_TYPO3);
        $formTitle = $this->formSpec[COLUMN_TITLE];

        // Construct the URL for the note button link
        $noteUrl = 'p:' . $pageSlug . '?form=' . NOTE_ITEM_NAME . '&formId=' . $formId . '&r=0&xId=' . $xId . '&formName=' . $formName . '&formTitle=' . $formTitle . '&type=1|s|r:7';


        $noteUrl = $this->link->renderLink($noteUrl);

        $targetForm = $formName . '-' . $recordId;

        // define the new window size
        $onClickJs = "window.open('$noteUrl', '$targetForm' , 'scrollbars=no,resizable=yes,width=1100,height=950'); return false;";

        // Build the anchor tag for the button with all the properties set
        $buttonNote = $this->buildButtonAnchor($noteUrl, NOTE_ITEM_BUTTON_HTML_ID, $this->formSpec[F_NOTE_BUTTON_TEXT], $toolTip, GLYPH_ICON_PAPERCLIP, $disabled, 'btn btn-default navbar-btn', $badge, $onClickJs);


        return Support::wrapTag('<div class="btn-group" role="group">', $buttonNote, true);
    }*/

    /**
     * Generic function to create a button with a given $buttonHtmlId, $url, $title, $icon (=glyph), $disabled
     *
     * @param string $url
     * @param string $buttonHtmlId
     * @param string $text
     * @param string $toolTip
     * @param string $icon
     * @param string $disabled
     * @param string $class
     *
     * @return string
     * @throws \CodeException
     */
    // TODO: Remove when not needed
    /*
    private function buildButtonAnchor($url, $buttonHtmlId, $text, $toolTip, $icon, $disabled = '', $class = '', $badge = '', $onClickJs = '') {


        $element = $text;

        // If there's an icon, prepend it to the element content
        if ($icon !== '') {
            $iconSpan = "<span " . Support::doAttribute('class', "glyphicon $icon") . "></span>";
            $element = $iconSpan . ' ' . $element;
        }
        // If there's a badge, append it to the element content
        if ($badge !== '') {
            $element = $element . ' ' . $badge;
        }

        $attribute = Support::doAttribute('href', $url);
        $attribute .= Support::doAttribute('id', $buttonHtmlId);
        $attribute .= Support::doAttribute('class', "$class $disabled");
        $attribute .= Support::doAttribute('title', $toolTip);

        // Conditionally add the `onclick` attribute if $onClickJs is not empty
        if (!empty($onClickJs)) {
            $attribute .= Support::doAttribute('onclick', $onClickJs);
        }

        // disabled links do not show tooltips -> use a span
        $wrapTag = $disabled == 'disabled' ? 'span' : 'a';

        return Support::wrapTag("<$wrapTag $attribute>", $element);
    }*/
}
